@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -10,14 +10,14 @@ import sys
 from xpra.client.client_base import XpraClientBase
 from xpra.client.keyboard_helper import KeyboardHelper
 from xpra.platform import set_name
-from xpra.platform.features import MMAP_SUPPORTED
 from xpra.platform.gui import ready as gui_ready, get_wm_name, get_session_type, ClientExtras
 from xpra.version_util import full_version_str
 from xpra.net import compression, packet_encoding
 from xpra.child_reaper import reaper_cleanup
 from xpra.os_util import platform_name, bytestostr, strtobytes, BITS, POSIX, is_Wayland
 from xpra.util import (
-    std, envbool, envint, typedict, updict, repr_ellipsized,
+    std, envbool, envint, typedict, updict, repr_ellipsized, ellipsizer, log_screen_sizes, engs, csv,
+    merge_dicts,
     XPRA_AUDIO_NOTIFICATION_ID, XPRA_DISCONNECT_NOTIFICATION_ID,
     )
 from xpra.exit_codes import EXIT_CONNECTION_FAILED, EXIT_OK, EXIT_CONNECTION_LOST
@@ -57,6 +57,9 @@ if mixin_features.logging:
 if mixin_features.network_state:
     from xpra.client.mixins.network_state import NetworkState
     CLIENT_BASES.append(NetworkState)
+if mixin_features.network_listener:
+    from xpra.client.mixins.network_listener import NetworkListener
+    CLIENT_BASES.append(NetworkListener)
 if mixin_features.encoding:
     from xpra.client.mixins.encodings import Encodings
     CLIENT_BASES.append(Encodings)
@@ -111,7 +114,7 @@ class UIXpraClient(ClientBaseClass):
 
         self._ui_events = 0
         self.title = ""
-        self.session_name = u""
+        self.session_name = ""
 
         self.server_platform = ""
         self.server_session_name = None
@@ -132,9 +135,7 @@ class UIXpraClient(ClientBaseClass):
         self.server_sharing_toggle = False
         self.server_lock = False
         self.server_lock_toggle = False
-        self.server_window_filters = False
         self.server_keyboard = True
-        self.server_toggle_keyboard_sync = False
         self.server_pointer = True
 
         self.client_supports_opengl = False
@@ -156,11 +157,11 @@ class UIXpraClient(ClientBaseClass):
         self._on_server_setting_changed = {}
 
 
-    def init(self, opts, extra_args=()):
+    def init(self, opts):
         """ initialize variables from configuration """
         for c in CLIENT_BASES:
             log("init: %s", c)
-            c.init(self, opts, extra_args)
+            c.init(self, opts)
 
         self.title = opts.title
         self.session_name = bytestostr(opts.session_name)
@@ -212,7 +213,7 @@ class UIXpraClient(ClientBaseClass):
                     #some platforms don't detect the vrefresh correctly
                     #(ie: macos in virtualbox?), so use a sane default:
                     v = 60
-                self._mouse_position_delay = 1000//v
+                self._mouse_position_delay = 1000//v//2
                 log("mouse delay: %s", self._mouse_position_delay)
             except Exception:
                 log("failed to calculate automatic delay", exc_info=True)
@@ -223,7 +224,6 @@ class UIXpraClient(ClientBaseClass):
             self.idle_add(self.client_extras.ready)
         for c in CLIENT_BASES:
             c.run(self)
-        self.send_hello()
 
 
     def quit(self, _exit_code=0):
@@ -254,6 +254,17 @@ class UIXpraClient(ClientBaseClass):
         log("UIXpraClient.signal_cleanup() done")
 
 
+    def get_info(self):
+        info = {}
+        for c in CLIENT_BASES:
+            try:
+                i = c.get_info(self)
+                info = merge_dicts(info, i)
+            except Exception:
+                log.error("Error collection information from %s", c, exc_info=True)
+        return info
+
+
     def show_about(self, *_args):
         log.warn("show_about() is not implemented in %s", self)
 
@@ -299,7 +310,7 @@ class UIXpraClient(ClientBaseClass):
         assert name is not None and command is not None and ignore is not None
         self.send("start-command", name, command, ignore, sharing)
 
-    def get_version_info(self):
+    def get_version_info(self) -> dict:
         return get_version_info_full()
 
 
@@ -309,7 +320,7 @@ class UIXpraClient(ClientBaseClass):
     def server_disconnect_warning(self, reason, *info):
         if self.exit_code is None:
             body = "\n".join(info)
-            if self.server_capabilities:
+            if self.connection_established:
                 title = "Xpra Session Disconnected: %s" % reason
                 self.exit_code = EXIT_CONNECTION_LOST
             else:
@@ -339,24 +350,24 @@ class UIXpraClient(ClientBaseClass):
         caps["session-type"] = get_session_type()
         #don't try to find the server uuid if this platform cannot run servers..
         #(doing so causes lockups on win32 and startup errors on osx)
-        if MMAP_SUPPORTED and POSIX and not is_Wayland():
+        if POSIX and not is_Wayland():
             #we may be running inside another server!
             try:
                 from xpra.server.server_uuid import get_uuid
                 caps["server_uuid"] = get_uuid() or ""
-            except:
+            except ImportError:
                 pass
         for x in (#generic feature flags:
-            "notify-startup-complete", "wants_events", "setting-change",
+            "wants_events", "setting-change",
             "xdg-menu-update",
             ):
             caps[x] = True
-        #FIXME: the messy bits without proper namespace:
         caps.update({
             #generic server flags:
             "share"                     : self.client_supports_sharing,
             "lock"                      : self.client_lock,
             })
+        caps.update({"mouse" : True})
         caps.update(self.get_keyboard_caps())
         for c in CLIENT_BASES:
             caps.update(c.get_caps(self))
@@ -364,6 +375,7 @@ class UIXpraClient(ClientBaseClass):
             updict(caps, prefix, c, flatten_dicts=False)
         u("control_commands",   self.get_control_commands_caps())
         u("platform",           get_platform_info())
+        u("opengl",             self.opengl_props)
         return caps
 
 
@@ -371,23 +383,25 @@ class UIXpraClient(ClientBaseClass):
     ######################################################################
     # connection setup:
     def setup_connection(self, conn):
+        protocol = super().setup_connection(conn)
         for c in CLIENT_BASES:
-            c.setup_connection(self, conn)
+            if c!=XpraClientBase:
+                c.setup_connection(self, conn)
+        return protocol
 
-    def server_connection_established(self):
-        if not XpraClientBase.server_connection_established(self):
+    def server_connection_established(self, caps : typedict):
+        if not XpraClientBase.server_connection_established(self, caps):
             return False
         #process the rest from the UI thread:
-        self.idle_add(self.process_ui_capabilities)
+        self.idle_add(self.process_ui_capabilities, caps)
         return True
 
 
-    def parse_server_capabilities(self):
-        for c in CLIENT_BASES:
-            if not c.parse_server_capabilities(self):
-                log.info("failed to parse server capabilities in %s", c)
-                return  False
-        c = self.server_capabilities
+    def parse_server_capabilities(self, c : typedict) -> bool:
+        for cb in CLIENT_BASES:
+            if not cb.parse_server_capabilities(self, c):
+                log.info("failed to parse server capabilities in %s", cb)
+                return False
         self.server_session_name = strtobytes(c.rawget("session_name", b"")).decode("utf-8")
         set_name("Xpra", self.session_name or self.server_session_name or "Xpra")
         self.server_platform = c.strget("platform")
@@ -396,7 +410,6 @@ class UIXpraClient(ClientBaseClass):
         self.server_lock = c.boolget("lock")
         self.server_lock_toggle = c.boolget("lock-toggle")
         self.server_keyboard = c.boolget("keyboard", True)
-        self.server_toggle_keyboard_sync = self.server_keyboard and c.boolget("toggle_keyboard_sync", True)
         self.server_pointer = c.boolget("pointer", True)
         self.server_start_new_commands = c.boolget("start-new-commands")
         if self.server_start_new_commands:
@@ -408,7 +421,7 @@ class UIXpraClient(ClientBaseClass):
                 log.warn("Warning: cannot start new commands")
                 log.warn(" the feature is currently disabled on the server")
         self.server_commands_info = c.boolget("server-commands-info")
-        self.server_commands_signals = c.strlistget("server-commands-signals")
+        self.server_commands_signals = c.strtupleget("server-commands-signals")
         self.server_readonly = c.boolget("readonly")
         if self.server_readonly and not self.readonly:
             log.info("server is read only")
@@ -420,7 +433,7 @@ class UIXpraClient(ClientBaseClass):
             self.keyboard_helper.send = nosend
 
         i = platform_name(self._remote_platform,
-                          c.strlistget("platform.linux_distribution") or c.strget("platform.release", ""))
+                          c.strtupleget("platform.linux_distribution") or c.strget("platform.release", ""))
         r = self._remote_version
         if self._remote_revision:
             r += "-r%s" % self._remote_revision
@@ -429,6 +442,16 @@ class UIXpraClient(ClientBaseClass):
         log.info("Xpra %s server version %s %i-bit", mode, std(r), bits)
         if i:
             log.info(" running on %s", std(i))
+        if c.boolget("desktop") or c.boolget("shadow"):
+            v = c.intpair("actual_desktop_size")
+            if v:
+                w, h = v
+                ss = c.tupleget("screen_sizes")
+                if ss:
+                    log.info(" remote desktop size is %sx%s with %s screen%s:", w, h, len(ss), engs(ss))
+                    log_screen_sizes(w, h, ss)
+                else:
+                    log.info(" remote desktop size is %sx%s", w, h)
         if c.boolget("proxy"):
             proxy_hostname = c.strget("proxy.hostname")
             proxy_platform = c.strget("proxy.platform")
@@ -445,17 +468,16 @@ class UIXpraClient(ClientBaseClass):
             log.info(msg)
         return True
 
-    def process_ui_capabilities(self):
+    def process_ui_capabilities(self, caps : typedict):
         for c in CLIENT_BASES:
             if c!=XpraClientBase:
-                c.process_ui_capabilities(self)
+                c.process_ui_capabilities(self, caps)
         #keyboard:
-        c = self.server_capabilities
         if self.keyboard_helper:
-            modifier_keycodes = c.dictget("modifier_keycodes", {})
+            modifier_keycodes = caps.dictget("modifier_keycodes", {})
             if modifier_keycodes:
                 self.keyboard_helper.set_modifier_mappings(modifier_keycodes)
-        self.key_repeat_delay, self.key_repeat_interval = c.intpair("key_repeat", (-1,-1))
+        self.key_repeat_delay, self.key_repeat_interval = caps.intpair("key_repeat", (-1,-1))
         self.handshake_complete()
 
 
@@ -478,7 +500,7 @@ class UIXpraClient(ClientBaseClass):
                 log.error("Error processing handshake callback %s", cb, exc_info=True)
 
     def after_handshake(self, cb, *args):
-        log("after_handshake(%s, %s) on_handshake=%s", cb, args, repr_ellipsized(str(self._on_handshake)))
+        log("after_handshake(%s, %s) on_handshake=%s", cb, args, ellipsizer(self._on_handshake))
         if self._on_handshake is None:
             #handshake has already occurred, just call it:
             self.idle_add(cb, *args)
@@ -489,7 +511,7 @@ class UIXpraClient(ClientBaseClass):
     ######################################################################
     # server messages:
     def _process_server_event(self, packet):
-        log(u": ".join((str(x) for x in packet[1:])))
+        log(": ".join((str(x) for x in packet[1:])))
 
     def on_server_setting_changed(self, setting, cb):
         self._on_server_setting_changed.setdefault(setting, []).append(cb)
@@ -563,23 +585,52 @@ class UIXpraClient(ClientBaseClass):
             #TODO: reset tray tooltip, session info title, etc..
         elif command=="debug":
             args = packet[2:]
-            if len(args)<2:
+            if not args:
                 log.warn("not enough arguments for debug control command")
                 return
+            from xpra.log import (
+                add_debug_category, add_disabled_category,
+                enable_debug_for, disable_debug_for,
+                get_all_loggers,
+                )
+            log_cmd = bytestostr(args[0])
+            if log_cmd=="status":
+                dloggers = [x for x in get_all_loggers() if x.is_debug_enabled()]
+                if dloggers:
+                    log.info("logging is enabled for:")
+                    for l in dloggers:
+                        log.info(" - %s", l)
+                else:
+                    log.info("logging is not enabled for any loggers")
+                return
             log_cmd = bytestostr(args[0])
             if log_cmd not in ("enable", "disable"):
                 log.warn("invalid debug control mode: '%s' (must be 'enable' or 'disable')", log_cmd)
                 return
-            categories = tuple(bytestostr(x) for x in args[1:])
-            from xpra.log import add_debug_category, add_disabled_category, enable_debug_for, disable_debug_for
-            if log_cmd=="enable":
-                add_debug_category(*categories)
-                loggers = enable_debug_for(*categories)
+            if len(args)<2:
+                log.warn("not enough arguments for '%s' debug control command" % log_cmd)
+                return
+            loggers = []
+            #each argument is a group
+            groups = [bytestostr(x) for x in args[1:]]
+            for group in groups:
+                #and each group is a list of categories
+                #preferably separated by "+",
+                #but we support "," for backwards compatibility:
+                categories = [v.strip() for v in group.replace("+", ",").split(",")]
+                if log_cmd=="enable":
+                    add_debug_category(*categories)
+                    loggers += enable_debug_for(*categories)
+                else:
+                    assert log_cmd=="disable"
+                    add_disabled_category(*categories)
+                    loggers += disable_debug_for(*categories)
+            if not loggers:
+                log.info("%s debugging, no new loggers matching: %s", log_cmd, csv(groups))
             else:
-                assert log_cmd=="disable"
-                add_disabled_category(*categories)
-                loggers = disable_debug_for(*categories)
-            log.info("%sd debugging for: %s", log_cmd, loggers)
+                log.info("%sd debugging for:", log_cmd)
+                for l in loggers:
+                    log.info(" - %s", l)
         else:
             log.warn("received invalid control command from server: %s", command)
 
@@ -616,7 +667,6 @@ class UIXpraClient(ClientBaseClass):
         self.send("force-ungrab", wid)
 
     def send_keyboard_sync_enabled_status(self, *_args):
-        assert self.server_toggle_keyboard_sync
         self.send("set-keyboard-sync-enabled", self.keyboard_sync)
 
 
