@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2017-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2017-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -8,15 +8,19 @@ import socket
 from time import sleep
 
 from xpra.scripts.config import InitException, TRUE_OPTIONS
+from xpra.scripts.main import InitExit
+from xpra.exit_codes import (
+    EXIT_SSL_FAILURE, EXIT_SSL_CERTIFICATE_VERIFY_FAILURE,
+    EXIT_SERVER_ALREADY_EXISTS, EXIT_SOCKET_CREATION_ERROR,
+    )
 from xpra.os_util import (
-    bytestostr, hexstr,
+    hexstr, bytestostr,
     getuid, get_username_for_uid, get_groups, get_group_id,
     path_permission_info, monotonic_time, umask_context, WIN32, OSX, POSIX,
     )
-from xpra.net.bytestreams import set_socket_timeout
 from xpra.util import (
-    envint, envbool, csv,
-    repr_ellipsized,
+    envint, envbool, csv, parse_simple_dict,
+    ellipsizer, repr_ellipsized,
     DEFAULT_PORT,
     )
 
@@ -25,6 +29,7 @@ WAIT_PROBE_TIMEOUT = envint("XPRA_WAIT_PROBE_TIMEOUT", 6)
 GROUP = os.environ.get("XPRA_GROUP", "xpra")
 PEEK_TIMEOUT = envint("XPRA_PEEK_TIMEOUT", 1)
 PEEK_TIMEOUT_MS = envint("XPRA_PEEK_TIMEOUT_MS", PEEK_TIMEOUT*1000)
+PEEK_SIZE = envint("XPRA_PEEK_SIZE", 8192)
 
 
 network_logger = None
@@ -46,7 +51,7 @@ def create_unix_domain_socket(sockpath, socket_permissions=0o600):
         listener.bind(sockpath)
     try:
         inode = os.stat(sockpath).st_ino
-    except (OSError, IOError):
+    except OSError:
         inode = -1
     #set to the "xpra" group if we are a member of it, or if running as root:
     uid = getuid()
@@ -68,7 +73,7 @@ def create_unix_domain_socket(sockpath, socket_permissions=0o600):
         log = get_network_logger()
         try:
             cur_inode = os.stat(sockpath).st_ino
-        except (OSError, IOError):
+        except OSError:
             log.info("socket '%s' already deleted", sockpath)
             return
         delpath = sockpath
@@ -77,11 +82,11 @@ def create_unix_domain_socket(sockpath, socket_permissions=0o600):
             log.info("removing socket '%s'", delpath)
             try:
                 os.unlink(delpath)
-            except (OSError, IOError):
+            except OSError:
                 pass
     return listener, cleanup_socket
 
-def has_dual_stack():
+def has_dual_stack() -> bool:
     """
         Return True if kernel allows creating a socket which is able to
         listen for both IPv4 and IPv6 connections.
@@ -124,20 +129,16 @@ def add_listen_socket(socktype, sock, info, new_connection_cb, new_udp_connectio
             assert new_udp_connection_cb, "UDP sockets cannot be handled here"
             new_udp_connection_cb(sock)
             return None
-        from xpra.gtk_common.gobject_compat import import_glib, is_gtk3
-        glib = import_glib()
+        from gi.repository import GLib
         sock.listen(5)
         def io_in_cb(sock, flags):
             log("io_in_cb(%s, %s)", sock, flags)
             return new_connection_cb(socktype, sock)
-        if is_gtk3():
-            source = glib.io_add_watch(sock, glib.PRIORITY_DEFAULT, glib.IO_IN, io_in_cb)
-        else:
-            source = glib.io_add_watch(sock, glib.IO_IN, io_in_cb, priority=glib.PRIORITY_DEFAULT)
+        source = GLib.io_add_watch(sock, GLib.PRIORITY_DEFAULT, GLib.IO_IN, io_in_cb)
         sources = [source]
         def cleanup():
             for source in tuple(sources):
-                glib.source_remove(source)
+                GLib.source_remove(source)
                 sources.remove(source)
         return cleanup
     except Exception as e:
@@ -147,7 +148,7 @@ def add_listen_socket(socktype, sock, info, new_connection_cb, new_udp_connectio
         return None
 
 
-def accept_connection(socktype, listener, timeout=None):
+def accept_connection(socktype, listener, timeout=None, socket_options=None):
     log = get_network_logger()
     try:
         sock, address = listener.accept()
@@ -159,39 +160,41 @@ def accept_connection(socktype, listener, timeout=None):
     #log("peercred(%s)=%s", sock, get_peercred(sock))
     try:
         peername = sock.getpeername()
-    except (OSError, IOError):
+    except OSError:
         peername = address
     sock.settimeout(timeout)
     sockname = sock.getsockname()
     from xpra.net.bytestreams import SocketConnection
-    conn = SocketConnection(sock, sockname, address, peername, socktype)
+    conn = SocketConnection(sock, sockname, address, peername, socktype, None, socket_options)
     log("accept_connection(%s, %s, %s)=%s", listener, socktype, timeout, conn)
     return conn
 
-def peek_connection(conn, timeout=PEEK_TIMEOUT_MS):
+def peek_connection(conn, timeout=PEEK_TIMEOUT_MS, size=PEEK_SIZE):
     log = get_network_logger()
     log("peek_connection(%s, %i)", conn, timeout)
-    PEEK_SIZE = 8192
-    start = monotonic_time()
     peek_data = b""
-    while not peek_data and int(1000*(monotonic_time()-start))<timeout:
+    start = monotonic_time()
+    elapsed = 0
+    while elapsed<=timeout:
         try:
-            peek_data = conn.peek(PEEK_SIZE)
-        except (OSError, IOError):
-            pass
+            peek_data = conn.peek(size)
+            if peek_data:
+                break
+        except OSError:
+            log("peek_connection(%s, %i) failed", conn, timeout, exc_info=True)
         except ValueError:
             log("peek_connection(%s, %i) failed", conn, timeout, exc_info=True)
             break
-        if not peek_data:
-            sleep(timeout/4000.0)
+        sleep(timeout/4000.0)
+        elapsed = int(1000*(monotonic_time()-start))
+        log("peek: elapsed=%s, timeout=%s", elapsed, timeout)
     line1 = b""
     log("socket %s peek: got %i bytes", conn, len(peek_data))
-    set_socket_timeout(conn, PEEK_TIMEOUT_MS*1000)
     if peek_data:
         line1 = peek_data.splitlines()[0]
-        log("socket peek=%s", repr_ellipsized(peek_data, limit=512))
+        log("socket peek=%s", ellipsizer(peek_data, limit=512))
         log("socket peek hex=%s", hexstr(peek_data[:128]))
-        log("socket peek line1=%s", repr_ellipsized(bytestostr(line1)))
+        log("socket peek line1=%s", ellipsizer(line1))
     return peek_data, line1
 
 
@@ -200,30 +203,30 @@ def create_sockets(opts, error_cb):
 
     bind_tcp = parse_bind_ip(opts.bind_tcp)
     bind_udp = parse_bind_ip(opts.bind_udp)
-    bind_ssl = parse_bind_ip(opts.bind_ssl)
-    bind_ssh = parse_bind_ip(opts.bind_ssh)
-    bind_ws  = parse_bind_ip(opts.bind_ws)
-    bind_wss = parse_bind_ip(opts.bind_wss)
+    bind_ssl = parse_bind_ip(opts.bind_ssl, 443)
+    bind_ssh = parse_bind_ip(opts.bind_ssh, 22)
+    bind_ws  = parse_bind_ip(opts.bind_ws, 80)
+    bind_wss = parse_bind_ip(opts.bind_wss, 443)
     bind_rfb = parse_bind_ip(opts.bind_rfb, 5900)
     bind_vsock = parse_bind_vsock(opts.bind_vsock)
 
-    sockets = []
+    sockets = {}
 
     min_port = int(opts.min_port)
-    def add_tcp_socket(socktype, host_str, iport):
+    def add_tcp_socket(socktype, host_str, iport, options):
         if iport!=0 and iport<min_port:
             error_cb("invalid %s port number %i (minimum value is %i)" % (socktype, iport, min_port))
         for host in hosts(host_str):
             sock = setup_tcp_socket(host, iport, socktype)
             host, iport = sock[2]
-            sockets.append(sock)
-    def add_udp_socket(socktype, host_str, iport):
+            sockets[sock] = options
+    def add_udp_socket(socktype, host_str, iport, options):
         if iport!=0 and iport<min_port:
             error_cb("invalid %s port number %i (minimum value is %i)" % (socktype, iport, min_port))
         for host in hosts(host_str):
             sock = setup_udp_socket(host, iport, socktype)
             host, iport = sock[2]
-            sockets.append(sock)
+            sockets[sock] = options
     # Initialize the TCP sockets before the display,
     # That way, errors won't make us kill the Xvfb
     # (which may not be ours to kill at that point)
@@ -238,34 +241,27 @@ def create_sockets(opts, error_cb):
             from xpra.log import Logger
             sshlog = Logger("ssh")
             sshlog("import paramiko", exc_info=True)
-            ssh_upgrades = False
             sshlog.error("Error: cannot enable SSH socket upgrades:")
             sshlog.error(" %s", e)
-    log("setting up SSL sockets: %s", csv(bind_ssl))
-    for host, iport in bind_ssl:
-        add_tcp_socket("ssl", host, iport)
-    log("setting up SSH sockets: %s", csv(bind_ssh))
-    for host, iport in bind_ssh:
-        add_tcp_socket("ssh", host, iport)
-    log("setting up https / wss (secure websockets): %s", csv(bind_wss))
-    for host, iport in bind_wss:
-        add_tcp_socket("wss", host, iport)
-    log("setting up TCP sockets: %s", csv(bind_tcp))
-    for host, iport in bind_tcp:
-        add_tcp_socket("tcp", host, iport)
-    log("setting up UDP sockets: %s", csv(bind_udp))
-    for host, iport in bind_udp:
-        add_udp_socket("udp", host, iport)
-    log("setting up http / ws (websockets): %s", csv(bind_ws))
-    for host, iport in bind_ws:
-        add_tcp_socket("ws", host, iport)
-    log("setting up rfb sockets: %s", csv(bind_rfb))
-    for host, iport in bind_rfb:
-        add_tcp_socket("rfb", host, iport)
-    log("setting up vsock sockets: %s", csv(bind_vsock))
-    for cid, iport in bind_vsock:
+            ssh_upgrades = False
+    for socktype, defs in {
+        "tcp"   : bind_tcp,
+        "ssl"   : bind_ssl,
+        "ssh"   : bind_ssh,
+        "ws"    : bind_ws,
+        "wss"   : bind_wss,
+        "rfb"   : bind_rfb,
+        }.items():
+        log("setting up %s sockets: %s", socktype, csv(defs.items()))
+        for (host, iport), options in defs.items():
+            add_tcp_socket(socktype, host, iport, options)
+    log("setting up UDP sockets: %s", csv(bind_udp.items()))
+    for (host, iport), options in bind_udp.items():
+        add_udp_socket("udp", host, iport, options)
+    log("setting up vsock sockets: %s", csv(bind_vsock.items()))
+    for (cid, iport), options in bind_vsock.items():
         sock = setup_vsock_socket(cid, iport)
-        sockets.append(sock)
+        sockets[sock] = options
 
     # systemd socket activation:
     if POSIX and not OSX:
@@ -277,7 +273,8 @@ def create_sockets(opts, error_cb):
             sd_sockets = get_sd_listen_sockets()
             log("systemd sockets: %s", sd_sockets)
             for stype, sock, addr in sd_sockets:
-                sockets.append(setup_sd_listen_socket(stype, sock, addr))
+                sock = setup_sd_listen_socket(stype, sock, addr)
+                sockets[sock] = {}
                 log("%s : %s", (stype, [addr]), sock)
     return sockets
 
@@ -305,12 +302,13 @@ def setup_tcp_socket(host, iport, socktype="tcp"):
         tcp_socket = create_tcp_socket(host, iport)
     except Exception as e:
         log("create_tcp_socket%s", (host, iport), exc_info=True)
-        raise InitException("failed to setup %s socket on %s:%s %s" % (socktype, host, iport, e))
+        raise InitExit(EXIT_SOCKET_CREATION_ERROR,
+                       "failed to setup %s socket on %s:%s %s" % (socktype, host, iport, e)) from None
     def cleanup_tcp_socket():
         log.info("closing %s socket '%s:%s'", socktype.lower(), host, iport)
         try:
             tcp_socket.close()
-        except (OSError, IOError):
+        except OSError:
             pass
     if iport==0:
         iport = tcp_socket.getsockname()[1]
@@ -337,27 +335,31 @@ def setup_udp_socket(host, iport, socktype="udp"):
         udp_socket = create_udp_socket(host, iport)
     except Exception as e:
         log("create_udp_socket%s", (host, iport), exc_info=True)
-        raise InitException("failed to setup %s socket on %s:%s %s" % (socktype, host, iport, e))
+        raise InitExit(EXIT_SOCKET_CREATION_ERROR,
+                       "failed to setup %s socket on %s:%s %s" % (socktype, host, iport, e))
     def cleanup_udp_socket():
         log.info("closing %s socket %s:%s", socktype, host, iport)
         try:
             udp_socket.close()
-        except (OSError, IOError):
+        except OSError:
             pass
     if iport==0:
         iport = udp_socket.getsockname()[1]
         log.info("allocated UDP port %i for %s", iport, host)
     log("%s: %s:%s : %s", socktype, host, iport, socket)
+    log.info("created UDP socket %s:%s", host, iport)
     return socktype, udp_socket, (host, iport), cleanup_udp_socket
 
 
 def parse_bind_ip(bind_ip, default_port=DEFAULT_PORT):
-    ip_sockets = set()
+    ip_sockets = {}
     if bind_ip:
         for spec in bind_ip:
+            parts = spec.split(",", 1)
+            ip_port = parts[0]
             if ":" not in spec:
                 raise InitException("port must be specified as [HOST]:PORT")
-            host, port = spec.rsplit(":", 1)
+            host, port = ip_port.rsplit(":", 1)
             if host == "":
                 host = "127.0.0.1"
             if not port:
@@ -367,10 +369,13 @@ def parse_bind_ip(bind_ip, default_port=DEFAULT_PORT):
             else:
                 try:
                     iport = int(port)
-                    assert iport>0 and iport<2**16
+                    assert 0<iport<2**16
                 except:
                     raise InitException("invalid port number: %s" % port)
-            ip_sockets.add((host, iport))
+            options = {}
+            if len(parts)==2:
+                options = parse_simple_dict(parts[1])
+            ip_sockets[(host, iport)] = options
     return ip_sockets
 
 def setup_vsock_socket(cid, iport):
@@ -379,21 +384,27 @@ def setup_vsock_socket(cid, iport):
         from xpra.net.vsock import bind_vsocket     #@UnresolvedImport
         vsock_socket = bind_vsocket(cid=cid, port=iport)
     except Exception as e:
-        raise InitException("failed to setup vsock socket on %s:%s %s" % (cid, iport, e))
+        raise InitExit(EXIT_SOCKET_CREATION_ERROR,
+                       "failed to setup vsock socket on %s:%s %s" % (cid, iport, e)) from None
     def cleanup_vsock_socket():
         log.info("closing vsock socket %s:%s", cid, iport)
         try:
             vsock_socket.close()
-        except (OSError, IOError):
+        except OSError:
             pass
     return "vsock", vsock_socket, (cid, iport), cleanup_vsock_socket
 
 def parse_bind_vsock(bind_vsock):
-    vsock_sockets = set()
+    vsock_sockets = {}
     if bind_vsock:
         from xpra.scripts.main import parse_vsock
         for spec in bind_vsock:
-            vsock_sockets.add(parse_vsock(spec))
+            parts = spec.split(",", 1)
+            cid, iport = parse_vsock(parts[0])
+            options = {}
+            if len(parts)==2:
+                options = parse_simple_dict(parts[1])
+            vsock_sockets[(cid, iport)] = options
     return vsock_sockets
 
 def setup_sd_listen_socket(stype, sock, addr):
@@ -402,7 +413,7 @@ def setup_sd_listen_socket(stype, sock, addr):
         log.info("closing sd listen socket %s", addr)
         try:
             sock.close()
-        except (OSError, IOError):
+        except OSError:
             pass
     return stype, sock, addr, cleanup_sd_listen_socket
 
@@ -414,6 +425,10 @@ def normalize_local_display_name(local_display_name):
         local_display_name = ":" + local_display_name
     else:
         after_sc = local_display_name[pos+1:]
+    if WIN32:
+        if after_sc.isalnum():
+            return local_display_name
+        raise Exception("non alphanumeric character in display name '%s'" % local_display_name)
     #we used to strip the screen from the display string, ie: ":0.0" -> ":0"
     #but now we allow it.. (untested!)
     for char in after_sc:
@@ -424,67 +439,74 @@ def normalize_local_display_name(local_display_name):
 def setup_local_sockets(bind, socket_dir, socket_dirs, display_name, clobber,
                         mmap_group="auto", socket_permissions="600", username="", uid=0, gid=0):
     if not bind:
-        return []
+        return {}
     if not socket_dir and (not socket_dirs or (len(socket_dirs)==1 and not socket_dirs[0])):
         if WIN32:
             socket_dirs = [""]
         else:
-            raise InitException("at least one socket directory must be set to use unix domain sockets")
+            raise InitExit(EXIT_SOCKET_CREATION_ERROR,
+                           "at least one socket directory must be set to use unix domain sockets")
     from xpra.platform.dotxpra import DotXpra, norm_makepath
     dotxpra = DotXpra(socket_dir or socket_dirs[0], socket_dirs, username, uid, gid)
     if display_name is not None:
         display_name = normalize_local_display_name(display_name)
     log = get_network_logger()
-    defs = []
+    defs = {}
     try:
-        sockpaths = []
+        sockpaths = {}
         log("setup_local_sockets: bind=%s", bind)
         for b in bind:
-            sockpath = b
             if b in ("none", ""):
                 continue
-            elif b=="auto":
+            parts = b.split(",")
+            sockpath = parts[0]
+            options = {}
+            if len(parts)==2:
+                options = parse_simple_dict(parts[1])
+            if sockpath=="auto":
                 assert display_name is not None
-                sockpaths += dotxpra.norm_socket_paths(display_name)
+                for sockpath in dotxpra.norm_socket_paths(display_name):
+                    sockpaths[sockpath] = options
                 log("sockpaths(%s)=%s (uid=%i, gid=%i)", display_name, sockpaths, uid, gid)
             else:
-                sockpath = dotxpra.osexpand(b)
-                if b.endswith("/") or (os.path.exists(sockpath) and os.path.isdir(sockpath)):
+                sockpath = dotxpra.osexpand(sockpath)
+                if os.path.isabs(sockpath):
+                    pass
+                elif sockpath.endswith("/") or (os.path.exists(sockpath) and os.path.isdir(sockpath)):
                     assert display_name is not None
                     sockpath = os.path.abspath(sockpath)
                     if not os.path.exists(sockpath):
                         os.makedirs(sockpath)
                     sockpath = norm_makepath(sockpath, display_name)
-                elif os.path.isabs(b):
-                    sockpath = b
                 else:
-                    sockpath = dotxpra.socket_path(b)
-                sockpaths += [sockpath]
+                    sockpath = dotxpra.socket_path(sockpath)
+                sockpaths[sockpath] = options
             assert sockpaths, "no socket paths to try for %s" % b
         #expand and remove duplicate paths:
-        tmp = []
-        for tsp in sockpaths:
+        tmp = {}
+        for tsp, options in sockpaths.items():
             sockpath = dotxpra.osexpand(tsp)
             if sockpath in tmp:
                 log.warn("Warning: skipping duplicate bind path %s", sockpath)
                 continue
-            tmp.append(sockpath)
+            tmp[sockpath] = options
         sockpaths = tmp
         #create listeners:
         if WIN32:
             from xpra.platform.win32.namedpipes.listener import NamedPipeListener
-            for sockpath in sockpaths:
+            for sockpath, options in sockpaths.items():
                 npl = NamedPipeListener(sockpath)
                 log.info("created named pipe '%s'", sockpath)
-                defs.append(("named-pipe", npl, sockpath, npl.stop))
+                defs[("named-pipe", npl, sockpath, npl.stop)] = options
         else:
             def checkstate(sockpath, state):
                 if state not in (DotXpra.DEAD, DotXpra.UNKNOWN):
                     if state==DotXpra.INACCESSIBLE:
                         raise InitException("An xpra server is already running at %s\n" % (sockpath,))
-                    raise InitException("You already have an xpra server running at %s\n"
-                         "  (did you want 'xpra upgrade'?)"
-                         % (sockpath,))
+                    raise InitExit(EXIT_SERVER_ALREADY_EXISTS,
+                                   "You already have an xpra server running at %s\n"
+                                   "  (did you want 'xpra upgrade'?)"
+                                   % (sockpath,))
             #remove exisiting sockets if clobber is set,
             #otherwise verify there isn't a server already running
             #and create the directories for the sockets:
@@ -518,7 +540,6 @@ def setup_local_sockets(bind, socket_dir, socket_dirs, display_name, clobber,
             log("sockets in unknown state: %s", unknown)
             if unknown:
                 #re-probe them using threads so we can do them in parallel:
-                from time import sleep
                 from xpra.make_thread import start_thread
                 threads = []
                 def timeout_probe(sockpath):
@@ -550,7 +571,7 @@ def setup_local_sockets(bind, socket_dir, socket_dirs, display_name, clobber,
                     try:
                         if os.path.exists(sockpath):
                             os.unlink(sockpath)
-                    except (OSError, IOError):
+                    except OSError:
                         pass
                 #socket permissions:
                 if mmap_group.lower() in TRUE_OPTIONS:
@@ -567,26 +588,25 @@ def setup_local_sockets(bind, socket_dir, socket_dirs, display_name, clobber,
                         assert 0<=sperms<=0o777, "invalid socket permission value %s" % oct(sperms)
                     except ValueError:
                         raise ValueError("invalid socket permissions "+
-                                         "(must be an octal number): '%s'" % socket_permissions)
+                                         "(must be an octal number): '%s'" % socket_permissions) from None
                 #now try to create all the sockets:
-                for sockpath in sockpaths:
+                for sockpath, options in sockpaths.items():
                     #create it:
                     try:
                         sock, cleanup_socket = create_unix_domain_socket(sockpath, sperms)
                         log.info("created unix domain socket '%s'", sockpath)
-                        defs.append(("unix-domain", sock, sockpath, cleanup_socket))
+                        defs[("unix-domain", sock, sockpath, cleanup_socket)] = options
                     except Exception as e:
                         handle_socket_error(sockpath, sperms, e)
                         del e
     except:
-        for socktype, sock, sockpath, cleanup_socket in defs:
+        for sock, cleanup_socket in defs:
             try:
                 cleanup_socket()
             except Exception as e:
                 log.error("Error cleaning up socket %s:", sock)
                 log.error(" %s", e)
                 del e
-        defs = []
         raise
     return defs
 
@@ -601,7 +621,7 @@ def handle_socket_error(sockpath, sperms, e):
             log.info(" %s does not exist", dirname)
         #only show extra information if the socket permissions
         #would have been accessible by the group:
-        elif POSIX and (sperms & 0o40):
+        if POSIX and (sperms & 0o40):
             uid = getuid()
             username = get_username_for_uid(uid)
             groups = get_groups(username)
@@ -621,7 +641,21 @@ def handle_socket_error(sockpath, sperms, e):
     else:
         log.error("Error: failed to create socket '%s':", sockpath)
         log.error(" %s", e)
-        raise InitException("failed to create socket %s" % sockpath)
+        raise InitExit(EXIT_SOCKET_CREATION_ERROR,
+                       "failed to create socket %s" % sockpath)
+
+
+def guess_header_protocol(v):
+    c = int(v[0])
+    s = bytestostr(v)
+    get_network_logger().debug("guess_header_protocol(%r) first char=%#x", repr_ellipsized(s), c)
+    if c==0x16:
+        return "ssl", "SSL packet?"
+    if s[:4]=="SSH-":
+        return "ssh", "SSH packet"
+    if len(s)>=3 and s.split(" ")[0] in ("GET", "POST"):
+        return "HTTP", "HTTP %s request" % s.split(" ")[0]
+    return None, "character %#x, not an xpra client?" % c
 
 
 #warn just once:
@@ -661,7 +695,7 @@ def mdns_publish(display_name, listen_on, text_dict=None):
         f_listen_on[(get_interface_index(host), port)] = (host, port)
     try:
         name = socket.gethostname()
-    except (IOError, OSError):
+    except OSError:
         name = "Xpra"
     if display_name and not (OSX or WIN32):
         name += " %s" % display_name
@@ -680,3 +714,196 @@ def mdns_publish(display_name, listen_on, text_dict=None):
         index += 1
         aps.append(MDNSPublishers(listen, sn, service_type=service_type, text_dict=d))
     return aps
+
+
+SSL_ATTRIBUTES = (
+    "cert", "key", "ca_certs", "ca_data",
+    "protocol",
+    "client_verify_mode", "server_verify_mode", "verify_flags",
+    "check_hostname", "server_hostname",
+    "options", "ciphers",
+    )
+
+def get_ssl_attributes(opts, server_side=True, overrides=None):
+    args = {
+        "server_side"   : server_side,
+        }
+    for attr in SSL_ATTRIBUTES:
+        ssl_attr = "ssl_%s" % attr          #ie: "ssl_ca_certs"
+        option = ssl_attr.replace("_", "-") #ie: "ssl-ca-certs"
+        v = (overrides or {}).get(option)
+        if v is None:
+            v = getattr(opts, ssl_attr)
+        args[attr] = v
+    return args
+
+def ssl_wrap_socket(sock, **kwargs):
+    fn = get_ssl_wrap_socket_fn(**kwargs)
+    return fn(sock)
+
+def get_ssl_wrap_socket_fn(cert=None, key=None, ca_certs=None, ca_data=None,
+                        protocol="TLSv1_2",
+                        client_verify_mode="optional", server_verify_mode="required", verify_flags="X509_STRICT",
+                        check_hostname=False, server_hostname=None,
+                        options="ALL,NO_COMPRESSION", ciphers="DEFAULT",
+                        server_side=True):
+    if server_side and not cert:
+        raise InitException("you must specify an 'ssl-cert' file to use ssl sockets")
+    if server_side:
+        verify_mode = client_verify_mode
+    else:
+        verify_mode = server_verify_mode
+    from xpra.log import Logger
+    ssllog = Logger("ssl")
+    ssllog("get_ssl_wrap_socket_fn%s", (cert, key, ca_certs, ca_data,
+                                    protocol,
+                                    client_verify_mode, server_verify_mode, verify_flags,
+                                    check_hostname, server_hostname,
+                                    options, ciphers,
+                                    server_side))
+    import ssl
+    ssllog(" verify_mode(%s)=%s", server_side, verify_mode)
+    #ca-certs:
+    if ca_certs=="default":
+        ca_certs = None
+    ssllog(" ca_certs=%s", ca_certs)
+    #parse verify-mode:
+    ssl_cert_reqs = getattr(ssl, "CERT_%s" % verify_mode.upper(), None)
+    if ssl_cert_reqs is None:
+        values = [k[len("CERT_"):].lower() for k in dir(ssl) if k.startswith("CERT_")]
+        raise InitException("invalid ssl-server-verify-mode '%s', must be one of: %s" % (verify_mode, csv(values)))
+    ssllog(" cert_reqs=%#x", ssl_cert_reqs)
+    #parse protocol:
+    proto = getattr(ssl, "PROTOCOL_%s" % (protocol.upper().replace("V", "v")), None)
+    if proto is None:
+        values = [k[len("PROTOCOL_"):] for k in dir(ssl) if k.startswith("PROTOCOL_")]
+        raise InitException("invalid ssl-protocol '%s', must be one of: %s" % (protocol, csv(values)))
+    ssllog(" protocol=%#x", proto)
+    #ca_data may be hex encoded:
+    if ca_data:
+        import binascii
+        try:
+            ca_data = binascii.unhexlify(ca_data)
+        except (TypeError, binascii.Error):
+            import base64
+            ca_data = base64.b64decode(ca_data)
+    ssllog(" cadata=%s", ellipsizer(ca_data))
+
+    kwargs = {
+              "server_side"             : server_side,
+              "do_handshake_on_connect" : False,
+              "suppress_ragged_eofs"    : True,
+              }
+    #parse ssl-verify-flags as CSV:
+    ssl_verify_flags = 0
+    for x in verify_flags.split(","):
+        x = x.strip()
+        if not x:
+            continue
+        v = getattr(ssl, "VERIFY_"+x.upper(), None)
+        if v is None:
+            raise InitException("invalid ssl verify-flag: %s" % x)
+        ssl_verify_flags |= v
+    ssllog(" verify_flags=%#x", ssl_verify_flags)
+    #parse ssl-options as CSV:
+    ssl_options = 0
+    for x in options.split(","):
+        x = x.strip()
+        if not x:
+            continue
+        v = getattr(ssl, "OP_"+x.upper(), None)
+        if v is None:
+            raise InitException("invalid ssl option: %s" % x)
+        ssl_options |= v
+    ssllog(" options=%#x", ssl_options)
+
+    context = ssl.SSLContext(proto)
+    context.set_ciphers(ciphers)
+    context.verify_mode = ssl_cert_reqs
+    context.verify_flags = ssl_verify_flags
+    context.options = ssl_options
+    ssllog(" cert=%s, key=%s", cert, key)
+    if cert:
+        SSL_KEY_PASSWORD = os.environ.get("XPRA_SSL_KEY_PASSWORD")
+        context.load_cert_chain(certfile=cert or None, keyfile=key or None, password=SSL_KEY_PASSWORD)
+    if ssl_cert_reqs!=ssl.CERT_NONE:
+        if server_side:
+            purpose = ssl.Purpose.CLIENT_AUTH   #@UndefinedVariable
+        else:
+            purpose = ssl.Purpose.SERVER_AUTH   #@UndefinedVariable
+            context.check_hostname = check_hostname
+            ssllog(" check_hostname=%s", check_hostname)
+            if context.check_hostname:
+                if not server_hostname:
+                    raise InitException("ssl error: check-hostname is set but server-hostname is not")
+                kwargs["server_hostname"] = server_hostname
+        ssllog(" load_default_certs(%s)", purpose)
+        context.load_default_certs(purpose)
+
+        if not ca_certs or ca_certs.lower()=="default":
+            ssllog(" loading default verify paths")
+            context.set_default_verify_paths()
+        elif not os.path.exists(ca_certs):
+            raise InitException("invalid ssl-ca-certs file or directory: %s" % ca_certs)
+        elif os.path.isdir(ca_certs):
+            ssllog(" loading ca certs from directory '%s'", ca_certs)
+            context.load_verify_locations(capath=ca_certs)
+        else:
+            ssllog(" loading ca certs from file '%s'", ca_certs)
+            assert os.path.isfile(ca_certs), "'%s' is not a valid ca file" % ca_certs
+            context.load_verify_locations(cafile=ca_certs)
+        #handle cadata:
+        if ca_data:
+            #PITA: because of a bug in the ssl module, we can't pass cadata,
+            #so we use a temporary file instead:
+            import tempfile
+            with tempfile.NamedTemporaryFile(prefix='cadata') as f:
+                ssllog(" loading cadata '%s'", ellipsizer(ca_data))
+                ssllog(" using temporary file '%s'", f.name)
+                f.file.write(ca_data)
+                f.file.flush()
+                context.load_verify_locations(cafile=f.name)
+    elif check_hostname and not server_side:
+        raise InitException("cannot check hostname with verify mode %s" % verify_mode)
+    wrap_socket = context.wrap_socket
+    def do_wrap_socket(tcp_socket):
+        assert tcp_socket
+        ssllog("do_wrap_socket(%s)", tcp_socket)
+        if WIN32:
+            #on win32, setting the tcp socket to blocking doesn't work?
+            #we still hit the following errors that we need to retry:
+            from xpra.net import bytestreams
+            bytestreams.CAN_RETRY_EXCEPTIONS = (ssl.SSLWantReadError, ssl.SSLWantWriteError)
+        else:
+            tcp_socket.setblocking(True)
+        try:
+            ssl_sock = wrap_socket(tcp_socket, **kwargs)
+        except Exception as e:
+            ssllog.debug("wrap_socket(%s, %s)", tcp_socket, kwargs, exc_info=True)
+            SSLEOFError = getattr(ssl, "SSLEOFError", None)
+            if SSLEOFError and isinstance(e, SSLEOFError):
+                return None
+            raise InitExit(EXIT_SSL_FAILURE, "Cannot wrap socket %s: %s" % (tcp_socket, e))
+        if not server_side:
+            try:
+                ssl_sock.do_handshake(True)
+            except Exception as e:
+                ssllog.debug("do_handshake", exc_info=True)
+                SSLEOFError = getattr(ssl, "SSLEOFError", None)
+                if SSLEOFError and isinstance(e, SSLEOFError):
+                    return None
+                status = EXIT_SSL_FAILURE
+                SSLCertVerificationError = getattr(ssl, "SSLCertVerificationError", None)
+                if SSLCertVerificationError and isinstance(e, SSLCertVerificationError):
+                    try:
+                        msg = e.args[1].split(":", 2)[2]
+                    except (ValueError, IndexError):
+                        msg = str(e)
+                    status = EXIT_SSL_CERTIFICATE_VERIFY_FAILURE
+                    #ssllog.warn("host failed SSL verification: %s", msg)
+                else:
+                    msg = str(e)
+                raise InitExit(status, "SSL handshake failed: %s" % msg)
+        return ssl_sock
+    return do_wrap_socket
+
