@@ -1,35 +1,38 @@
 # This file is part of Xpra.
-# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2019-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import os
 from ctypes import (
     sizeof, byref, cast,
     get_last_error, create_string_buffer,
     WinError, FormatError,
     )
+from gi.repository import GLib
+
 from xpra.platform.win32.common import (
-    WNDCLASSEX, GetLastError, WNDPROC, LPCWSTR, LPWSTR,
+    WNDCLASSEX, GetLastError,
+    WNDPROC, LPCWSTR, LPWSTR, LPCSTR, DWORD,
     DefWindowProcW,
     GetModuleHandleA, RegisterClassExW, UnregisterClassA,
     CreateWindowExW, DestroyWindow,
-    OpenClipboard, EmptyClipboard, CloseClipboard, GetClipboardData, GetClipboardOwner,
+    OpenClipboard, EmptyClipboard, CloseClipboard, GetClipboardData,
     GlobalLock, GlobalUnlock, GlobalAlloc, GlobalFree,
     WideCharToMultiByte, MultiByteToWideChar,
     AddClipboardFormatListener, RemoveClipboardFormatListener,
-    SetClipboardData)
+    SetClipboardData, EnumClipboardFormats, GetClipboardFormatNameA, GetClipboardOwner,
+    GetWindowThreadProcessId, QueryFullProcessImageNameA, OpenProcess, CloseHandle,
+    )
 from xpra.platform.win32 import win32con
 from xpra.clipboard.clipboard_timeout_helper import ClipboardTimeoutHelper
 from xpra.clipboard.clipboard_core import (
     ClipboardProxyCore, log, _filter_targets,
     TEXT_TARGETS, MAX_CLIPBOARD_PACKET_SIZE,
     )
-from xpra.util import csv, repr_ellipsized, envbool
+from xpra.util import csv, ellipsizer, envint, envbool
 from xpra.os_util import bytestostr, strtobytes
-from xpra.gtk_common.gobject_compat import import_glib
-
-glib = import_glib()
-
+from xpra.platform.win32.constants import PROCESS_QUERY_INFORMATION
 
 CP_UTF8 = 65001
 MB_ERR_INVALID_CHARS = 0x00000008
@@ -55,19 +58,65 @@ CLIPBOARD_EVENTS = {
     win32con.WM_VSCROLLCLIPBOARD    : "WM_VSCROLLCLIPBOARD",
     }
 
-CONVERT_LINE_ENDINGS = envbool("XPRA_CONVERT_LINE_ENDINGS", True)
+CLIPBOARD_FORMATS = {
+    win32con.CF_BITMAP      : "CF_BITMAP",
+    win32con.CF_DIB         : "CF_DIB",
+    win32con.CF_DIBV5       : "CF_DIBV5",
+    win32con.CF_ENHMETAFILE : "CF_ENHMETAFILE",
+    win32con.CF_METAFILEPICT: "CF_METAFILEPICT",
+    win32con.CF_OEMTEXT     : "CF_OEMTEXT",
+    win32con.CF_TEXT        : "CF_TEXT",
+    win32con.CF_UNICODETEXT : "CF_UNICODETEXT",
+    }
 
+RETRY = envint("XPRA_CLIPBOARD_RETRY", 5)
+DELAY = envint("XPRA_CLIPBOARD_INITIAL_DELAY", 10)
+CONVERT_LINE_ENDINGS = envbool("XPRA_CONVERT_LINE_ENDINGS", True)
+log("win32 clipboard: RETRY=%i, DELAY=%i, CONVERT_LINE_ENDINGS=%s",
+    RETRY, DELAY, CONVERT_LINE_ENDINGS)
+#can be used to blacklist problematic clipboard peers:
+#ie: VBoxTray.exe
+BLACKLISTED_CLIPBOARD_CLIENTS = [x for x in
+                                 os.environ.get("XPRA_BLACKLISTED_CLIPBOARD_CLIENTS", "").split(",")
+                                 if x]
+log("BLACKLISTED_CLIPBOARD_CLIENTS=%s", BLACKLISTED_CLIPBOARD_CLIENTS)
+
+
+def is_blacklisted(owner_info):
+    return any(owner_info.find(x)>=0 for x in BLACKLISTED_CLIPBOARD_CLIENTS)
 
 #initialize the window we will use
 #for communicating with the OS clipboard API:
 
+def get_owner_info(owner, our_window):
+    if not owner:
+        return "unknown"
+    if owner==our_window:
+        return "our window (hwnd=%#x)" % our_window
+    pid = DWORD(0)
+    GetWindowThreadProcessId(owner, byref(pid))
+    if not pid:
+        return "unknown (hwnd=%#x)" % owner
+    #log("get_owner_info(%#x) pid=%s", owner, pid.value)
+    proc_handle = OpenProcess(PROCESS_QUERY_INFORMATION, False, pid)
+    if not proc_handle:
+        return "pid %i (hwnd=%#x)" % (pid.value, owner)
+    try:
+        size = DWORD(256)
+        process_name = create_string_buffer(size.value+1)
+        if not QueryFullProcessImageNameA(proc_handle, 0, process_name, byref(size)):
+            return "pid %i" % pid.value
+        return "'%s' with pid %s (hwnd=%#x)" % (bytestostr(process_name.value), pid.value, owner)
+    finally:
+        CloseHandle(proc_handle)
+
 class Win32Clipboard(ClipboardTimeoutHelper):
     """
         Use Native win32 API to access the clipboard
     """
     def __init__(self, send_packet_cb, progress_cb=None, **kwargs):
         self.init_window()
-        ClipboardTimeoutHelper.__init__(self, send_packet_cb, progress_cb, **kwargs)
+        super().__init__(send_packet_cb, progress_cb, **kwargs)
 
     def init_window(self):
         log("Win32Clipboard.init_window() creating clipboard window class and instance")
@@ -83,10 +132,10 @@ class Win32Clipboard(ClipboardTimeoutHelper):
         if self.wndclass_handle==0:
             raise WinError()
         style = win32con.WS_CAPTION   #win32con.WS_OVERLAPPED
-        self.window = CreateWindowExW(0, self.wndclass_handle, u"Clipboard", style,
+        self.window = CreateWindowExW(0, self.wndclass_handle, "Clipboard", style,
                                       0, 0, win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT,
                                       win32con.HWND_MESSAGE, 0, self.wndclass.hInstance, None)
-        log("clipboard window=%s", self.window)
+        log("clipboard window=%#x", self.window)
         if not self.window:
             raise WinError()
         if not AddClipboardFormatListener(self.window):
@@ -96,8 +145,16 @@ class Win32Clipboard(ClipboardTimeoutHelper):
     def wnd_proc(self, hwnd, msg, wparam, lparam):
         r = DefWindowProcW(hwnd, msg, wparam, lparam)
         if msg in CLIPBOARD_EVENTS:
-            log("clipboard event: %s", CLIPBOARD_EVENTS.get(msg))
-        if msg==WM_CLIPBOARDUPDATE:
+            owner = GetClipboardOwner()
+            log("clipboard event: %s, current owner: %s",
+                CLIPBOARD_EVENTS.get(msg), get_owner_info(owner, self.window))
+        if msg==WM_CLIPBOARDUPDATE and owner!=self.window:
+            owner = GetClipboardOwner()
+            owner_info = get_owner_info(owner, self.window)
+            if is_blacklisted(owner_info):
+                #ie: don't try to sync from VirtualBox
+                log("CLIPBOARDUPDATE coming from '%s' ignored", owner_info)
+                return r
             for proxy in self._clipboard_proxies.values():
                 if not proxy._block_owner_change:
                     proxy.schedule_emit_token()
@@ -134,13 +191,13 @@ class Win32Clipboard(ClipboardTimeoutHelper):
     def _munge_wire_selection_to_raw(self, encoding, dtype, dformat, data):
         if encoding=="atoms":
             return _filter_targets(data)
-        return ClipboardTimeoutHelper._munge_wire_selection_to_raw(self, encoding, dtype, dformat, data)
+        return super()._munge_wire_selection_to_raw(encoding, dtype, dformat, data)
 
     def _munge_raw_selection_to_wire(self, target, dtype, dformat, data):
         if dtype=="ATOM":
             assert isinstance(data, (tuple, list))
             return "atoms", _filter_targets(data)
-        return ClipboardTimeoutHelper._munge_raw_selection_to_wire(self, target, dtype, dformat, data)
+        return super()._munge_raw_selection_to_wire(target, dtype, dformat, data)
 
 
 class Win32ClipboardProxy(ClipboardProxyCore):
@@ -148,31 +205,44 @@ class Win32ClipboardProxy(ClipboardProxyCore):
         self.window = window
         self.send_clipboard_request_handler = send_clipboard_request_handler
         self.send_clipboard_token_handler = send_clipboard_token_handler
-        ClipboardProxyCore.__init__(self, selection)
+        super().__init__(selection)
 
-    def with_clipboard_lock(self, success_callback, failure_callback, retries=5, delay=5):
+    def with_clipboard_lock(self, success_callback, failure_callback, retries=RETRY, delay=DELAY):
+        log("with_clipboard_lock%s", (success_callback, failure_callback, retries, delay))
         r = OpenClipboard(self.window)
         if r:
             log("OpenClipboard(%#x)=%s", self.window, r)
             try:
-                success_callback()
-                return
+                r = success_callback()
+                log("%s()=%s", success_callback, r)
+                if r:
+                    return
             finally:
-                CloseClipboard()
-        log("OpenClipboard(%#x)=%s, owner=%#x", self.window, WinError(GetLastError()), GetClipboardOwner())
+                r = CloseClipboard()
+                log("CloseClipboard()=%s", r)
+        e = WinError(GetLastError())
+        owner = GetClipboardOwner()
+        log("OpenClipboard(%#x)=%s, current owner: %s", self.window, e, get_owner_info(owner, self.window))
         if retries<=0:
-            failure_callback("OpenClipboard: too many failed attemps, giving up")
+            failure_callback("OpenClipboard: too many failed attempts, giving up")
             return
         #try again later:
-        glib.timeout_add(delay, self.with_clipboard_lock,
+        GLib.timeout_add(delay, self.with_clipboard_lock,
                          success_callback, failure_callback, retries-1, delay+5)
 
     def clear(self):
-        def clear_error(error_text=""):
-            log.error("Error: failed to clear the clipboard")
-            if error_text:
-                log.error(" %s", error_text)
-        self.with_clipboard_lock(EmptyClipboard, clear_error)
+        self.with_clipboard_lock(self.empty_clipboard, self.clear_error)
+
+    def empty_clipboard(self):
+        r = EmptyClipboard()
+        log("EmptyClipboard()=%s", r)
+        return True
+
+    def clear_error(self, error_text=""):
+        log.warn("Warning: failed to clear the clipboard")
+        if error_text:
+            log.warn(" %s", error_text)
+
 
     def do_emit_token(self):
         #TODO: if contents are not text,
@@ -183,8 +253,7 @@ class Win32ClipboardProxy(ClipboardProxyCore):
                 packet_data = ([target], (target, dtype, dformat, data))
                 self.send_clipboard_token_handler(self, packet_data)
             self.get_contents(target, got_contents)
-        else:
-            self.send_clipboard_token_handler(self)
+        self.send_clipboard_token_handler(self)
 
     def get_contents(self, target, got_contents):
         log("get_contents%s", (target, got_contents))
@@ -198,14 +267,16 @@ class Win32ClipboardProxy(ClipboardProxyCore):
             got_contents(target, 8, b"")
             return
         def got_text(text):
-            log("got_text(%s)", repr_ellipsized(bytestostr(text)))
+            log("got_text(%s)", ellipsizer(text))
             got_contents(target, 8, text)
         def errback(error_text=""):
-            log.error("Error: failed to get clipboard data")
+            log("errback(%s)", error_text)
             if error_text:
-                log.error(" %s", error_text)
+                log.warn("Warning: failed to get clipboard data")
+                log.warn(" %s", error_text)
             got_contents(target, 8, b"")
-        self.get_clipboard_text(got_text, errback)
+        utf8 = target.lower().find("utf")>=0
+        self.get_clipboard_text(utf8, got_text, errback)
 
 
     def got_token(self, targets, target_data=None, claim=True, _synchronous_client=False):
@@ -249,37 +320,89 @@ class Win32ClipboardProxy(ClipboardProxyCore):
             self.set_clipboard_text(data)
 
 
-    def get_clipboard_text(self, callback, errback):
+    def get_clipboard_text(self, utf8, callback, errback):
+        def format_name(fmt):
+            name = CLIPBOARD_FORMATS.get(fmt)
+            if name:
+                return name
+            ulen = 128
+            buf = LPCSTR(b" "*ulen)
+            r = GetClipboardFormatNameA(fmt, buf, ulen-1)
+            if r==0:
+                return str(fmt)
+            return (buf.value[:r]).decode("latin1")
         def get_text():
-            data_handle = GetClipboardData(win32con.CF_UNICODETEXT)
+            formats = []
+            fmt = 0
+            while True:
+                fmt = EnumClipboardFormats(fmt)
+                if fmt:
+                    formats.append(fmt)
+                else:
+                    break
+            log("clipboard formats: %s", csv(format_name(x) for x in formats))
+            matching = []
+            for u in (utf8, not utf8):
+                if u:
+                    fmts = [win32con.CF_UNICODETEXT]
+                else:
+                    fmts = [win32con.CF_TEXT, win32con.CF_OEMTEXT]
+                matching += [fmt for fmt in formats if fmt in fmts]
+            log("supported formats: %s (prefer utf8: %s)", csv(format_name(x) for x in matching), utf8)
+            if not matching:
+                log("no supported formats, only: %s", csv(format_name(x) for x in formats))
+                errback()
+                return True
+            data_handle = None
+            for fmt in matching:
+                data_handle = GetClipboardData(fmt)
+                log("GetClipboardData(%s)=%#x", format_name(fmt), data_handle or 0)
+                if data_handle:
+                    break
             if not data_handle:
-                errback("no data handle")
-                return
+                log("no valid data handle using %s (may try again)", csv(format_name(x) for x in matching))
+                return False
             data = GlobalLock(data_handle)
             if not data:
-                errback("failed to lock handle")
-                return
+                log("failed to lock data handle %#x (may try again)", data_handle)
+                return False
+            log("got data handle lock %#x for format '%s'", data, format_name(fmt))
             try:
-                wstr = cast(data, LPCWSTR)
-                ulen = WideCharToMultiByte(CP_UTF8, 0, wstr, -1, None, 0, None, None)
-                if ulen>MAX_CLIPBOARD_PACKET_SIZE:
-                    errback("too much data")
-                    return
-                buf = create_string_buffer(ulen)
-                l = WideCharToMultiByte(CP_UTF8, 0, wstr, -1, byref(buf), ulen, None, None)
-                if l>0:
+                if fmt==win32con.CF_UNICODETEXT:
+                    wstr = cast(data, LPCWSTR)
+                    ulen = WideCharToMultiByte(CP_UTF8, 0, wstr, -1, None, 0, None, None)
+                    if ulen>MAX_CLIPBOARD_PACKET_SIZE:
+                        errback("unicode data is too large: %i bytes" % ulen)
+                        return True
+                    buf = create_string_buffer(ulen)
+                    l = WideCharToMultiByte(CP_UTF8, 0, wstr, -1, byref(buf), ulen, None, None)
+                    if l==0:
+                        errback("failed to convert to UTF8: %s" % FormatError(get_last_error()))
+                        return True
                     if buf.raw[l-1:l]==b"\0":
                         s = buf.raw[:l-1]
                     else:
                         s = buf.raw[:l]
+                    log("got %i bytes of UNICODE data: %s", len(s), ellipsizer(s))
                     if CONVERT_LINE_ENDINGS:
                         v = s.decode("utf8").replace("\r\n", "\n").encode("utf8")
                     else:
                         v = strtobytes(s)
-                    log("got %i bytes of data: %s", len(s), repr_ellipsized(str(s)))
                     callback(v)
-                else:
-                    errback("failed to convert to UTF8: %s" % FormatError(get_last_error()))
+                    return True
+                #CF_TEXT or CF_OEMTEXT:
+                astr = cast(data, LPCSTR)
+                s = astr.value.decode("latin1")
+                if CONVERT_LINE_ENDINGS:
+                    s = s.replace("\r\n", "\n")
+                b = s.encode("latin1")
+                ulen = len(b)
+                if ulen>MAX_CLIPBOARD_PACKET_SIZE:
+                    errback("text data is too large: %i characters" % ulen)
+                    return True
+                log("got %i bytes of TEXT data: %s", len(b), ellipsizer(b))
+                callback(b)
+                return True
             finally:
                 GlobalUnlock(data)
         self.with_clipboard_lock(get_text, errback)
@@ -288,17 +411,7 @@ class Win32ClipboardProxy(ClipboardProxyCore):
         log.warn("Warning: cannot set clipboard value")
         log.warn(" %s", msg)
 
-    def set_clipboard_text(self, text, retry=5):
-        log("set_clipboard_text(%s, %i)", text, retry)
-        r = self.do_set_clipboard_text(text)
-        if not r:
-            if retry>0:
-                glib.timeout_add(5, self.set_clipboard_text, text, retry-1)
-            else:
-                self.set_err("failed to set clipboard buffer")
-
-
-    def do_set_clipboard_text(self, text):
+    def set_clipboard_text(self, text):
         #convert to wide char
         #get the length in wide chars:
         if CONVERT_LINE_ENDINGS:
@@ -306,54 +419,55 @@ class Win32ClipboardProxy(ClipboardProxyCore):
         wlen = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, text, len(text), None, 0)
         if not wlen:
             self.set_err("failed to prepare to convert to wide char")
-            return False
+            return True
         log("MultiByteToWideChar wlen=%i", wlen)
         #allocate some memory for it:
         l = (wlen+1)*2
         buf = GlobalAlloc(GMEM_MOVEABLE, l)
         if not buf:
             self.set_err("failed to allocate %i bytes of global memory" % l)
-            return False
+            return True
         log("GlobalAlloc buf=%#x", buf)
         locked = GlobalLock(buf)
         if not locked:
             self.set_err("failed to lock buffer %#x" % buf)
             GlobalFree(buf)
-            return False
+            return True
         try:
             locked_buf = cast(locked, LPWSTR)
             r = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, text, len(text), locked_buf, wlen)
             if not r:
                 self.set_err("failed to convert to wide char")
-                return False
+                GlobalFree(buf)
+                return True
         finally:
             GlobalUnlock(locked)
         #we're going to alter the clipboard ourselves,
         #ignore messages until we're done:
         self._block_owner_change = True
-        #def empty_error():
-        #    self.set_err("failed to empty the clipboard")
-        #self.with_clipboard_lock(EmptyClipboard, empty_error)
         def cleanup():
-            glib.idle_add(self.remove_block)
-        ret = [False]
-        def do_set_data():
-            if not EmptyClipboard():
+            GLib.idle_add(self.remove_block)
+        def set_clipboard_data():
+            r = EmptyClipboard()
+            log("EmptyClipboard()=%s", r)
+            if not r:
                 self.set_err("failed to empty the clipboard")
-            if not SetClipboardData(win32con.CF_UNICODETEXT, buf):
-                #no need to warn here
-                #set_clipboard_text() will try again
-                return
-            log("SetClipboardData(..) done")
+                return False
+            r = SetClipboardData(win32con.CF_UNICODETEXT, buf)
+            if not r:
+                e = WinError(GetLastError())
+                log("SetClipboardData(CF_UNICODETEXT, %i chars)=%s (%s)", wlen, r, e)
+                return False
+            log("SetClipboardData(CF_UNICODETEXT, %i chars)=%s", wlen, r)
             cleanup()
-            ret[0] = True
-        def set_error(error_text=""):
-            log.error("Error: failed to set clipboard data")
+            return True
+        def set_clipboard_error(error_text=""):
+            log("set_clipboard_error(%s)", error_text)
             if error_text:
-                log.error(" %s", error_text)
+                log.warn("Warning: failed to set clipboard data")
+                log.warn(" %s", error_text)
             cleanup()
-        self.with_clipboard_lock(do_set_data, set_error)
-        return ret[0]
+        self.with_clipboard_lock(set_clipboard_data, set_clipboard_error)
 
 
     def __repr__(self):
