@@ -12,40 +12,26 @@ import os
 import sys
 import glob
 from io import BytesIO
+from typing import Generator as generator       #@UnresolvedImport, @UnusedImport
 
 from xpra.util import envbool, envint, print_nested_dict, first_time
-from xpra.os_util import load_binary_file, OSEnvContext, PYTHON3
+from xpra.os_util import load_binary_file, OSEnvContext
 from xpra.log import Logger, add_debug_category
 
 log = Logger("exec", "menu")
 
 LOAD_GLOB = envbool("XPRA_XDG_LOAD_GLOB", True)
 EXPORT_ICONS = envbool("XPRA_XDG_EXPORT_ICONS", True)
-MAX_ICON_SIZE = envint("XPRA_XDG_MAX_ICON_SIZE", 0)
+MAX_ICON_SIZE = envint("XPRA_XDG_MAX_ICON_SIZE", 65536)
 DEBUG_COMMANDS = os.environ.get("XPRA_XDG_DEBUG_COMMANDS", "").split(",")
-if PYTHON3:
-    unicode = str           #@ReservedAssignment
-    from typing import Generator as generator       #@UnresolvedImport, @UnusedImport
-else:
-    from types import GeneratorType as generator    #@Reimport
 
-def check_xdg():
-    try:
-        from xdg.Menu import Menu, MenuEntry
-        assert Menu and MenuEntry
-        return True
-    except ImportError as e:
-        if first_time("load-xdg"):
-            log.warn("Warning: cannot load menu data")
-            log.warn(" %s", e)
-        return False
 
 def isvalidtype(v):
     if isinstance(v, (list, tuple, generator)):
         if not v:
             return True
         return all(isvalidtype(x) for x in v)
-    return isinstance(v, (bytes, str, unicode, bool, int))
+    return isinstance(v, (bytes, str, bool, int))
 
 def export(entry, properties):
     name = entry.getName()
@@ -86,10 +72,10 @@ def load_icon_from_file(filename):
     log("load_icon_from_file(%s)", filename)
     if filename.endswith("xpm"):
         try:
-            from xpra.gtk_common.gobject_compat import import_pixbufloader
             from xpra.gtk_common.gtk_util import pixbuf_save_to_memory
             data = load_binary_file(filename)
-            loader = import_pixbufloader()()
+            from gi.repository import GdkPixbuf
+            loader = GdkPixbuf.PixbufLoader()
             loader.write(data)
             loader.close()
             pixbuf = loader.get_pixbuf()
@@ -116,37 +102,8 @@ def load_icon_from_file(filename):
     icondata = load_binary_file(filename)
     if not icondata:
         return None
-    if filename.endswith("svg") and len(icondata)>MAX_ICON_SIZE//2:
-        #try to resize it
-        try:
-            size = len(icondata)
-            import cairo
-            import gi
-            try:
-                gi.require_version('Rsvg', '2.0')
-            except ValueError as e:
-                if first_time("no-rsvg"):
-                    log.warn("Warning: cannot resize svg icons, Rsvg bindings not found:")
-                    log.warn(" %s", e)
-            else:
-                from gi.repository import Rsvg
-                img = cairo.ImageSurface(cairo.FORMAT_ARGB32, 128, 128)
-                ctx = cairo.Context(img)
-                handle = Rsvg.Handle.new_from_data(icondata)
-                handle.render_cairo(ctx)
-                buf = BytesIO()
-                img.write_to_png(buf)
-                icondata = buf.getvalue()
-                buf.close()
-                log("reduced size of SVG icon %s, from %i bytes to %i bytes as PNG",
-                         filename, size, len(icondata))
-                filename = filename[:-3]+"png"
-        except ImportError:
-            log("cannot convert svg", exc_info=True)
-        except Exception:
-            log.error("Error: failed to convert svg icon", exc_info=True)
     log("got icon data from '%s': %i bytes", filename, len(icondata))
-    if MAX_ICON_SIZE>0 and len(icondata)>MAX_ICON_SIZE and first_time("icon-size-warning-%s" % filename):
+    if len(icondata)>MAX_ICON_SIZE and first_time("icon-size-warning-%s" % filename):
         log.warn("Warning: icon is quite large (%i KB):", len(icondata)//1024)
         log.warn(" '%s'", filename)
     return icondata, os.path.splitext(filename)[1].lstrip(".")
@@ -161,7 +118,7 @@ def load_icon_from_theme(icon_name, theme=None):
     return load_icon_from_file(filename)
 
 def load_glob_icon(submenu_data, main_dirname="categories"):
-    if not LOAD_GLOB or not EXPORT_ICONS:
+    if not LOAD_GLOB:
         return None
     #doesn't work with IconTheme.getIconPath,
     #so do it the hard way:
@@ -215,9 +172,6 @@ def load_xdg_entry(de):
             command = de.getTryExec()
     else:
         command = de.getExec()
-    if not command:
-        #this command is not executable!
-        return None
     props["command"] = command
     icondata = props.get("IconData")
     if not icondata:
@@ -280,8 +234,6 @@ def load_xdg_menu_data(force_reload=False):
     return xdg_menu_data
 
 def do_load_xdg_menu_data():
-    if not check_xdg():
-        return {}
     try:
         from xdg.Menu import parse, Menu, ParsingError
     except ImportError:
@@ -311,12 +263,10 @@ def do_load_xdg_menu_data():
                 try:
                     menu = parse()
                     break
-                except Exception as e:
+                except ParsingError as e:
                     log("do_load_xdg_menu_data()", exc_info=True)
                     error = e
                     menu = None
-            if menu:
-                break
     if menu is None:
         if error:
             log.error("Error parsing xdg menu data:")
