@@ -57,7 +57,6 @@ class InputServer(StubServerMixin):
 
     def get_server_features(self, _source=None):
         return {
-            "toggle_keyboard_sync"  : True,
             "input-devices"         : self.input_devices,
             "pointer.relative"      : True,
             }
@@ -87,14 +86,14 @@ class InputServer(StubServerMixin):
             self.key_repeat = c.intpair("key_repeat", (0, 0))
             self.set_keyboard_repeat(self.key_repeat)
             #always clear modifiers before setting a new keymap
-            ss.make_keymask_match(c.strlistget("modifiers", []))
+            ss.make_keymask_match(c.strtupleget("modifiers"))
         else:
             self.set_keyboard_repeat(None)
             self.key_repeat = (0, 0)
         self.key_repeat_delay, self.key_repeat_interval = self.key_repeat
         self.set_keymap(ss)
 
-    def get_keyboard_info(self):
+    def get_keyboard_info(self) -> dict:
         start = monotonic_time()
         info = {
              "repeat"           : {
@@ -294,8 +293,7 @@ class InputServer(StubServerMixin):
     def _keys_changed(self, *_args):
         if not self.keymap_changing:
             for ss in self._server_sources.values():
-                if hasattr(ss, "keys_changed"):
-                    ss.keys_changed()
+                ss.keys_changed()
 
     def clear_keys_pressed(self):
         pass
@@ -401,12 +399,12 @@ class InputServer(StubServerMixin):
         pass
 
 
+    #FIXME: we should not be overriding this method here
     def send_hello(self, server_source, _root_w, _root_h, key_repeat, _server_cipher):
         capabilities = self.make_hello(server_source)
         if key_repeat:
             capabilities.update({
                 "key_repeat"           : key_repeat,
-                "key_repeat_modifiers" : True,
                 })
 
 
