@@ -6,6 +6,7 @@
 
 import socket
 from time import sleep, time
+from queue import Queue
 
 from xpra.net import compression
 from xpra.net.net_util import get_network_caps
@@ -16,7 +17,7 @@ from xpra.codecs.image_wrapper import ImageWrapper
 from xpra.codecs.video_helper import getVideoHelper, PREFERRED_ENCODER_ORDER
 from xpra.scripts.config import parse_number, parse_bool
 from xpra.os_util import (
-    Queue, get_hex_uuid,
+    get_hex_uuid,
     monotonic_time, bytestostr, strtobytes,
     )
 from xpra.util import (
@@ -32,11 +33,11 @@ log = Logger("proxy")
 enclog = Logger("encoding")
 
 
-PROXY_QUEUE_SIZE = envint("XPRA_PROXY_QUEUE_SIZE", 0)
+PROXY_QUEUE_SIZE = envint("XPRA_PROXY_QUEUE_SIZE", 10)
 #for testing only: passthrough as RGB:
 PASSTHROUGH_RGB = envbool("XPRA_PROXY_PASSTHROUGH_RGB", False)
 VIDEO_TIMEOUT = 5                  #destroy video encoder after N seconds of idle state
-LEGACY_SALT_DIGEST = envbool("XPRA_LEGACY_SALT_DIGEST", False)
+LEGACY_SALT_DIGEST = envbool("XPRA_LEGACY_SALT_DIGEST", True)
 PASSTHROUGH_AUTH = envbool("XPRA_PASSTHROUGH_AUTH", True)
 
 PING_INTERVAL = max(1, envint("XPRA_PROXY_PING_INTERVAL", 5))*1000
@@ -47,7 +48,7 @@ CLIENT_REMOVE_CAPS = ("cipher", "challenge", "digest", "aliases", "compression",
 CLIENT_REMOVE_CAPS_CHALLENGE = ("cipher", "digest", "aliases", "compression", "lz4", "lz0", "zlib")
 
 
-class ProxyInstance(object):
+class ProxyInstance:
 
     def __init__(self, session_options,
                  video_encoder_modules, pings,
@@ -201,7 +202,7 @@ class ProxyInstance(object):
             "window" : self.get_window_info(),
             }
 
-    def get_window_info(self):
+    def get_window_info(self) -> dict:
         info = {}
         now = monotonic_time()
         for wid, encoder in self.video_encoders.items():
@@ -456,11 +457,11 @@ class ProxyInstance(object):
             if self.cipher:
                 from xpra.net.crypto import crypto_backend_init, new_cipher_caps, DEFAULT_PADDING
                 crypto_backend_init()
-                padding_options = self.caps.strlistget("cipher.padding.options", [DEFAULT_PADDING])
+                padding_options = self.caps.strtupleget("cipher.padding.options", [DEFAULT_PADDING])
                 auth_caps = new_cipher_caps(self.client_protocol, self.cipher, self.encryption_key, padding_options)
                 caps.update(auth_caps)
             #may need to bump packet size:
-            proto.max_packet_size = max(16*1024*1024, maxw*maxh*4*4)
+            proto.max_packet_size = maxw*maxh*4*4
             file_transfer = self.caps.boolget("file-transfer") and c.boolget("file-transfer")
             file_size_limit = max(self.caps.intget("file-size-limit"), c.intget("file-size-limit"))
             file_max_packet_size = int(file_transfer) * (1024 + file_size_limit*1024*1024)
@@ -491,11 +492,6 @@ class ProxyInstance(object):
             self.encode_queue.put(packet)
             #which will queue the packet itself when done:
             return
-        elif packet_type=="sound-data":
-            sound_data = packet[2]
-            if sound_data:
-                #best if we use raw packets for the actual sound-data chunk:
-                packet[2] = Compressed("sound-data", sound_data)
         #we do want to reformat cursor packets...
         #as they will have been uncompressed by the network layer already:
         elif packet_type=="cursor":
@@ -574,15 +570,6 @@ class ProxyInstance(object):
 
     def encode_loop(self):
         """ thread for slower encoding related work """
-        def delvideo(wid):
-            try:
-                del self.video_encoders[wid]
-            except KeyError:
-                pass
-            try:
-                del self.video_encoders_last_used_time[wid]
-            except KeyError:
-                pass
         while not self.exit:
             packet = self.encode_queue.get()
             if packet is None:
@@ -594,7 +581,8 @@ class ProxyInstance(object):
                     self.lost_windows.remove(wid)
                     ve = self.video_encoders.get(wid)
                     if ve:
-                        delvideo(wid)
+                        del self.video_encoders[wid]
+                        del self.video_encoders_last_used_time[wid]
                         ve.clean()
                 elif packet_type==b"draw":
                     #modify the packet with the video encoder:
@@ -606,13 +594,14 @@ class ProxyInstance(object):
                     wid = packet[1]
                     ve = self.video_encoders.get(wid)
                     now = monotonic_time()
-                    idle_time = now-self.video_encoders_last_used_time.get(wid, 0)
+                    idle_time = now-self.video_encoders_last_used_time.get(wid)
                     if ve and idle_time>VIDEO_TIMEOUT:
                         enclog("timing out the video encoder context for window %s", wid)
                         #timeout is confirmed, we are in the encoding thread,
                         #so it is now safe to clean it up:
                         ve.clean()
-                        delvideo(wid)
+                        del self.video_encoders[wid]
+                        del self.video_encoders_last_used_time[wid]
                 else:
                     enclog.warn("unexpected encode packet: %s", packet_type)
             except Exception:
@@ -692,7 +681,7 @@ class ProxyInstance(object):
                 ve.clean()
                 ve = None
         #scaling and depth are proxy-encoder attributes:
-        scaling = client_options.intlistget("scaling", (1, 1))
+        scaling = client_options.inttupleget("scaling", (1, 1))
         depth   = client_options.intget("depth", 24)
         rowstride = client_options.intget("rowstride", rowstride)
         quality = client_options.intget("quality", -1)
@@ -723,7 +712,7 @@ class ProxyInstance(object):
             enclog("creating new video encoder %s for window %s", spec, wid)
             ve = spec.make_instance()
             #dst_formats is specified with first frame only:
-            dst_formats = client_options.strlistget("dst_formats")
+            dst_formats = client_options.strtupleget("dst_formats")
             if dst_formats is not None:
                 #save it in case we timeout the video encoder,
                 #so we can instantiate it again, even from a frame no>1
@@ -752,10 +741,9 @@ class ProxyInstance(object):
         #to do what may need to be done if we find a timeout:
         now = monotonic_time()
         for wid in tuple(self.video_encoders_last_used_time.keys()):
-            vetime = self.video_encoders_last_used_time.get(wid)
-            if not vetime:
+            idle_time = int(now-self.video_encoders_last_used_time.get(wid))
+            if idle_time is None:
                 continue
-            idle_time = int(now-vetime)
             enclog("timeout_video_encoders() wid=%s, idle_time=%s", wid, idle_time)
             if idle_time and idle_time>VIDEO_TIMEOUT:
                 self.encode_queue.put(["check-video-timeout", wid])
